﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Text.RegularExpressions;

namespace DialogueEditor
{
    static class DTEConverter
    {
        private static string[] dte = 
            { "<EOC>", "\r\n", "<A0>", "<A1>", "<A2>", "<A3>", "<A4>", "<A5>", "<A6>", "<A7>", "<A8>", "<A9>", "<A10>", "<A11>", "<A12>", "<A13>",
        "<D>","<E$xx>", "<OP$12>","<EOP>","<S$xx>","<C>","<D$xx>","<OP$17>","<OP$18>","<N>","<I>","<S>","","","","",
        "A","B","C","D","E","F","G","H","I","J","K","L","M","N","O","P",
        "Q","R","S","T","U","V","W","X","Y","Z","a","b","c","d","e","f",
        "g","h","i","j","k","l","m","n","o","p","q","r","s","t","u","v",
        "w","x","y","z","0","1","2","3","4","5","6","7","8","9","!","?",
        "/",":","]","'","-",".",",","_",";","#","+","(",")","%","~","*",
        "@","&","=","[","","","$","\\","^","","","{","}","`","|"," ",
        "e "," t",": ","th","t ","he","s ","er"," a","re","in","ou","d "," w"," s","an",
        "o "," h", " o", "r ","n ","at","to"," i",", ", "ve","ng","ha"," m","Th","st","on",
        "yo"," b","me","y ","en","it","ar","ll","ea","I ","ed"," f"," y","hi","is","es",
        "or","l "," c", "ne","'s","nd","le","se"," I","a ","te"," l","pe","as","ur","u ",
        "al"," p","g ","om"," d","f "," g","ow","rs","be","ro","us","ri","wa","we","Wh",
        "et"," r","nt","m ","ma","I'","li","ho","of","Yo","h "," n","ee","de","so","gh",
        "ca","ra","n'","ta","ut","el","! ","fo","ti","We","lo","e!","ld","no","ac","ce",
        "k "," u","oo","ke","ay","w ","!!","ag","il","ly","co",". ","ch","go","ge","e_"};

        public static string[] Names = { "TERRA", "LOCKE", "CYAN", "SHADOW", "EDGAR", "SABIN", "CELES", "STRAGO", "RELM", "SETZER", "MOG", "GAU", "GOGO", "UMARO" };

        public static int IndexAdvance;

        public static string StringFromByte(byte[] data, int index)
        {
            IndexAdvance = 1;
            string ret = dte[data[index]];
            if (ret.Contains("$xx"))
            {
                ret = ret.Replace("xx", data[index + 1].ToString("X2"));
                IndexAdvance = 2;
            }
            return ret;
        }

        public static byte[] BytesFromString(string instr)
        {
            if (instr != "")
            {
                string unoptimizedbytes = "";
                for (int i = 0; i < instr.Length; i++)
                {
                    string refstr = instr[i].ToString();
                    string paystr = ",";
                    if (refstr == "<")
                    {
                        while (instr[i] != '>')
                        {
                            i++;
                            refstr += instr[i];
                            if (instr[i] == '$' && refstr != "<OP$")
                            {
                                paystr += instr[i + 1].ToString() + instr[i + 2].ToString() + ",";
                                refstr += "xx";
                                i += 2;
                            }
                        }
                    }
                    if (refstr == "\r")
                    {
                        refstr = "\r\n";
                        i++;
                    }
                    //refstr = Array.IndexOf(dte, refstr) == -1 ? "\r\n" : refstr;
                    if (Array.IndexOf(dte, refstr) < 0)
                    {
                        return new byte[] { 0xFF };
                    }
                    else
                    {
                        unoptimizedbytes += Array.IndexOf(dte, refstr).ToString("X2") + paystr;
                    }
                }
                string[] ubytes = unoptimizedbytes.Substring(0, unoptimizedbytes.Length - 1).Split(',');
                string optimizedbytes = "";
                for (int j = 0; j < ubytes.Length; j++)
                {
                    string opt = ubytes[j];
                    if (j < ubytes.Length - 1)
                    {
                        string teststr = dte[int.Parse(ubytes[j], System.Globalization.NumberStyles.AllowHexSpecifier)] + dte[int.Parse(ubytes[j + 1], System.Globalization.NumberStyles.AllowHexSpecifier)];
                        if (Array.IndexOf(dte, teststr) >= 0x80)
                        {
                            opt = Array.IndexOf(dte, teststr).ToString("X2");
                            j++;
                        }
                    }
                    //catch { }
                    optimizedbytes += opt + ",";
                }
                optimizedbytes = !optimizedbytes.EndsWith("00,") ? optimizedbytes + "00," : optimizedbytes;
                string[] obytes = optimizedbytes.Substring(0, optimizedbytes.Length - 1).Split(',');
                return Array.ConvertAll(obytes, item => byte.Parse(item, System.Globalization.NumberStyles.AllowHexSpecifier));
            }
            else
            {
                return new byte[] { 0 }; 
            }
        }
    }
}
