﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;

namespace DialogueEditor
{
    public partial class DialogueEditor : Form
    {
        public byte[] data;
        public bool headered;
        public byte[] header;
        private int[] pivotindices;
        public int[] startbanks;
        public int[] maxoffsets;
        private List<DialogueItem> DialogueLines;
        private int maxindex;
        public List<int> AvailableBanks;
        private int rembytes;
        private int remcaptions;
        public bool isROMExpanded;
        public bool isDialogExpanded;
        public string ProgName;

        private string loadedfilename;

        public DialogueEditor()
        {
            InitializeComponent();
            DialogueLines = new List<DialogueItem>();
            pivotindices = new int[3];
            startbanks = new int[3];
            maxoffsets = new int[3];
            maxindex = 0x0CFF;
            this.KeyPreview = true;
            this.KeyDown += new KeyEventHandler(HotKeyEvent);
            AvailableBanks = new List<int>();
            rembytes = 0;
            remcaptions = 0;
            loadedfilename = "";
            isROMExpanded = false;
            isDialogExpanded = false;
            headered = false;
            ProgName = System.Reflection.Assembly.GetExecutingAssembly().GetName().Name.ToString() + " - " +
                System.Reflection.Assembly.GetExecutingAssembly().GetName().Version.ToString();
            this.Text = ProgName;
        }

        private void HotKeyEvent(object sender, KeyEventArgs e)
        {
            if (e.Control && e.KeyCode == Keys.S)       // Ctrl-S Save
            {
                BTN_Apply_Click(this, null);
                e.SuppressKeyPress = true;
            }
        }

        private void saveROMToolStripMenuItem_Click(object sender, EventArgs e)
        {
            WriteToData();
            byte [] savedata;
            if(headered)
            {
                savedata = new byte[data.Length + 0x200];
                for(int i = 0; i < header.Length; i++)
                {
                    savedata[i] = header[i];
                }
                for(int j = 0; j < data.Length; j++)
                {
                    savedata[0x200 + j] = data[j];
                }
            }
            else
            {
                savedata = data;
            }
            File.WriteAllBytes(loadedfilename, savedata);
            NeedsSave(false);
        }

        private void loadROMToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFileDialog ofd = new OpenFileDialog();
            ofd.ShowDialog();
            if (ofd.FileName != "")
            {
                headered = false;
                loadedfilename = ofd.FileName;
                data = File.ReadAllBytes(ofd.FileName);
                if(data[0x10] == 0)
                {
                    headered = true;
                    header = new byte[0x200];
                    for(int i = 0; i < header.Length; i++)
                    {
                        header[i] = data[i];
                    }
                    byte[] tempdata = new byte[data.Length - 0x200];
                    for(int j = 0; j < tempdata.Length; j++)
                    {
                        tempdata[j] = data[0x200 + j];
                    }
                    data = tempdata;
                }

                expandROMToolStripMenuItem.Enabled = true;

                if (data.Length > 0x300000)
                {
                    isROMExpanded = true;
                    expandROMToolStripMenuItem.Enabled = false;
                    expandROMToolStripMenuItem.Text = "ROM Already Expanded";
                    expandDialogueSpaceToolStripMenuItem.Enabled = true;
                }
                if (data[0xDF80] != 0xFF)
                {
                    isDialogExpanded = true;
                    expandDialogueSpaceToolStripMenuItem.Enabled = false;
                    expandDialogueSpaceToolStripMenuItem.Text = "Dialogue Already Expanded";
                    manageMemoryToolStripMenuItem.Enabled = true;
                }

                ReadFromData();
                saveROMToolStripMenuItem.Enabled = true;
                saveROMAsToolStripMenuItem.Enabled = true;
                BTN_Add.Enabled = true;
                BTN_Export.Enabled = true;
                BTN_Import.Enabled = true;
                BTN_Apply.Enabled = true;
            }

        }

        public void ReadFromData()
        {
            DialogueLines.Clear();
            int pointerstartaddr;
            if (isDialogExpanded)
            {
                pointerstartaddr = int.Parse(data[0x7FDA].ToString("X2") + data[0x7FD9].ToString("X2") + data[0x7FD8].ToString("X2"), System.Globalization.NumberStyles.AllowHexSpecifier) - 0xC00000;
                for (int b = 0; b < pivotindices.Length; b++)
                {
                    int addr = 0xDF80 + (b * 3);
                    pivotindices[b] = int.Parse(data[addr + 1].ToString("X2") + data[addr].ToString("X2"), System.Globalization.NumberStyles.AllowHexSpecifier);
                    startbanks[b] = int.Parse((data[addr + 2] - 0xC0).ToString("X2") + "0000", System.Globalization.NumberStyles.AllowHexSpecifier);
                    maxoffsets[b] = startbanks[b] == 0x0E0000 ? 0xF0FF : 0xFFFF;
                }
            }
            else
            {
                pointerstartaddr = 0xCE602;
                int addr = 0xCE600;
                pivotindices[0] = int.Parse(data[addr + 1].ToString("X2") + data[addr].ToString("X2"), System.Globalization.NumberStyles.AllowHexSpecifier);
                pivotindices[1] = 0xFFFF;
                pivotindices[2] = 0xFFFF;
                startbanks[0] = 0x0D0000;
                startbanks[1] = 0x0E0000;
                startbanks[2] = 0x300000;
                maxoffsets[0] = 0xFFFF;
                maxoffsets[1] = 0xF0FF;
                maxoffsets[2] = 0;
            }
            int bank = 0;
            //int dlgstartaddr = 0x0D0000;
            int idx = 1;
            int prevptr = 0xFFFF;
            bool endofpointers = false;
            rembytes = maxoffsets[0] + maxoffsets[1] + maxoffsets[2];
            remcaptions = (startbanks[0] == 0xD0000 || startbanks[1] == 0xD0000 || startbanks[2] == 0xD0000) ? 0x0CFF : 0x3FFF;
            while (!endofpointers)
            {
                int ptraddr = pointerstartaddr + (idx * 2);
                int dlgptr = int.Parse((data[ptraddr + 1].ToString("X2") + data[ptraddr].ToString("X2")), System.Globalization.NumberStyles.AllowHexSpecifier);
                if (dlgptr != prevptr || idx < 3082)
                {
                    prevptr = dlgptr;
                    int dlgbytecount = 0;
                    DialogueItem di = new DialogueItem(data[startbanks[bank] + dlgptr + dlgbytecount]);
                    remcaptions--;
                    dlgbytecount++;
                    rembytes--;
                    if (data[startbanks[bank] + dlgptr + dlgbytecount - 1] != 0)
                    {
                        byte prevbyte = 0;
                        while (data[startbanks[bank] + dlgptr + dlgbytecount] != 0 && !(data[startbanks[bank] + dlgptr + dlgbytecount] == 0xFF && prevbyte == 0xFF))
                        {
                            prevbyte = data[startbanks[bank] + dlgptr + dlgbytecount];
                            di.AddByte(data[startbanks[bank] + dlgptr + dlgbytecount]);
                            dlgbytecount++;
                            rembytes--;
                        }
                        di.AddByte(0);
                        rembytes--;
                    }
                    DialogueLines.Add(di);
                    idx++;
                    if (idx == 3236)
                    {

                    }
                    if (idx >= pivotindices[bank])
                    {
                        bank++;
                    }
                }
                else
                {
                    endofpointers = true;
                    DialogueLines.RemoveAt(DialogueLines.Count - 1);
                    remcaptions++;
                    LBL_RemBytes.Text = rembytes.ToString();
                    LBL_RemPointers.Text = remcaptions.ToString();
                }
            }

            LB_Dialog.Items.Clear();
            for (int i = 0; i < DialogueLines.Count; i++)
            {
                string lbtxt = i.ToString("0000") + ". ($" + (i + 1).ToString("X4") + "). " + DialogueLines[i].StrippedText;
                LB_Dialog.Items.Add(lbtxt);
            }
        }

        public void CheckAvailableBanks()
        {
            int startbank = 0;
            int i = 0;
            AvailableBanks.Clear();
            while (startbank < data.Length)
            {
                i = startbank;
                if (data[startbank] == 0xFF)
                {
                    bool available = true;
                    int endaddr = startbank == 0x0E0000 ? 0xF0FF : 0xFFFF;
                    while (i < startbank + endaddr)
                    {
                        i++;
                        if (data[i] != 0xFF)
                        {
                            available = false;
                            break;
                        }
                    }
                    if (available)
                    {
                        AvailableBanks.Add(startbank);
                    }
                }
                startbank += 0x10000;
            }
        }

        private void LB_Dialog_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (LB_Dialog.SelectedIndex >= 0)
            {
                RTB_Text.Text = DialogueLines[LB_Dialog.SelectedIndex].Text;
                int len = DialogueLines[LB_Dialog.SelectedIndex].Bytes.Length;
            }
        }

        private void saveROMAsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SaveFileDialog sfd = new SaveFileDialog();
            sfd.ShowDialog();
            if (sfd.FileName != "")
            {
                WriteToData();
                byte[] savedata;
                if (headered)
                {
                    savedata = new byte[data.Length + 0x200];
                    for (int i = 0; i < header.Length; i++)
                    {
                        savedata[i] = header[i];
                    }
                    for (int j = 0; j < data.Length; j++)
                    {
                        savedata[0x200 + j] = data[j];
                    }
                }
                else
                {
                    savedata = data;
                }
                File.WriteAllBytes(sfd.FileName, savedata);
                NeedsSave(false);
            }
        }

        public void WriteToData()
        {
            int pointerstartaddr;
            if (isDialogExpanded)
            {
                for (int j = 0; j < startbanks.Length; j++)
                {
                    int addr = 0xDF80 + (j * 3);
                    data[addr + 2] = byte.Parse((startbanks[j] + 0xC00000).ToString("X6").Substring(0, 2), System.Globalization.NumberStyles.AllowHexSpecifier);
                }
                pointerstartaddr = int.Parse(data[0x7FDA].ToString("X2") + data[0x7FD9].ToString("X2") + data[0x7FD8].ToString("X2"), System.Globalization.NumberStyles.AllowHexSpecifier) - 0xC00000;

            }
            else
            {
                pointerstartaddr = 0xCE602;
                int addr = 0xCE600;                
                startbanks[0] = 0x0D0000;
                startbanks[1] = 0x0E0000;
                maxoffsets[0] = 0xFFFF;
                maxoffsets[1] = 0xF0FF;
                maxoffsets[2] = 0;
            }

            int bank = 0;

            int idx = 1;
            int offset = 0;

            int ptraddr = 0;
            pivotindices[0] = 0xFFFF;
            pivotindices[1] = 0xFFFF;
            pivotindices[2] = 0xFFFF;
            foreach (DialogueItem di in DialogueLines)
            {
                ptraddr = pointerstartaddr + (idx * 2);
                int newoffset = offset;
                if (offset + di.Bytes.Length >= maxoffsets[bank])
                {
                    pivotindices[bank] = idx;
                    if (isDialogExpanded)
                    {
                        for (int f = offset; f <= maxoffsets[bank]; f++)
                        {
                            data[startbanks[bank] + f] = 0xFF;
                        }
                        offset = 0;
                        newoffset = 0;
                    }
                    else
                    {
                        newoffset = -(maxoffsets[bank] - offset + 1);
                        pivotindices[bank] = pivotindices[bank] + 1;
                    }
                    bank++;
                }
                byte[] offsetbytes = { byte.Parse(offset.ToString("X4").Substring(0, 2), System.Globalization.NumberStyles.AllowHexSpecifier), byte.Parse(offset.ToString("X4").Substring(2), System.Globalization.NumberStyles.AllowHexSpecifier) };
                offset = newoffset;
                data[ptraddr] = offsetbytes[1];
                data[ptraddr + 1] = offsetbytes[0];
                foreach (byte b in di.Bytes)
                {
                    data[startbanks[bank] + offset] = b;

                    offset++;
                }
                idx++;
            }
            data[startbanks[bank] + offset] = 0;

            maxindex = startbanks[0] != 0x0D0000 && startbanks[1] != 0x0D0000 & startbanks[2] != 0x0D0000 ? 0x3FFF : 0x0CFF;
            while (idx < maxindex)
            {
                ptraddr = pointerstartaddr + (idx * 2);
                byte[] offsetbytes = { byte.Parse(offset.ToString("X4").Substring(0, 2), System.Globalization.NumberStyles.AllowHexSpecifier), byte.Parse(offset.ToString("X4").Substring(2), System.Globalization.NumberStyles.AllowHexSpecifier) };

                data[ptraddr] = offsetbytes[1];
                data[ptraddr + 1] = offsetbytes[0];
                idx++;
            }
            if (isDialogExpanded)
            {
                for (int b = 0; b < pivotindices.Length; b++)
                {
                    int addr = 0xDF80 + (b * 3);
                    byte[] pivotbytes = { byte.Parse(pivotindices[b].ToString("X4").Substring(0, 2), System.Globalization.NumberStyles.AllowHexSpecifier), byte.Parse(pivotindices[b].ToString("X4").Substring(2), System.Globalization.NumberStyles.AllowHexSpecifier) };
                    data[addr] = pivotbytes[1];
                    data[addr + 1] = pivotbytes[0];
                }
            }
            else
            {
                int addr = 0xCE600;
                byte[] pivotbytes = { byte.Parse(pivotindices[0].ToString("X4").Substring(0, 2), System.Globalization.NumberStyles.AllowHexSpecifier), byte.Parse(pivotindices[0].ToString("X4").Substring(2), System.Globalization.NumberStyles.AllowHexSpecifier) };
                data[addr] = pivotbytes[1];
                data[addr + 1] = pivotbytes[0];
            }
            
        }

        public void EraseData(int startaddr, int endaddr)
        {
            for (int i = startaddr; i <= endaddr; i++)
            {
                data[i] = 0xFF;
            }
        }

        private void BTN_Apply_Click(object sender, EventArgs e)
        {
            byte[] bytecheck = DTEConverter.BytesFromString(DialogueLines[LB_Dialog.SelectedIndex].FormatTextForMemory(RTB_Text.Text));
            if (bytecheck.Length == 1 && bytecheck[0] == 0xFF)
            {
                MessageBox.Show("Error in message formatting. Text not saved.");
            }
            else
            {
                int origtextlen = DialogueLines[LB_Dialog.SelectedIndex].Bytes.Length;
                DialogueLines[LB_Dialog.SelectedIndex].Text = RTB_Text.Text;
                string lbtxt = LB_Dialog.SelectedIndex.ToString("0000") + ". ($" + (LB_Dialog.SelectedIndex + 1).ToString("X4") + "). " + DialogueLines[LB_Dialog.SelectedIndex].StrippedText;
                LB_Dialog.Items[LB_Dialog.SelectedIndex] = lbtxt;
                rembytes = rembytes - (DialogueLines[LB_Dialog.SelectedIndex].Bytes.Length - origtextlen);
                LBL_RemBytes.Text = rembytes.ToString();
                NeedsSave(true);
            }
        }

        private void manageMemoryToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ManageMemory mm = new ManageMemory(this);
            mm.ShowDialog();
        }

        private void BTN_Add_Click(object sender, EventArgs e)
        {
            if (remcaptions > 0 && rembytes > 0)
            {
                DialogueItem di = new DialogueItem(0);
                DialogueLines.Add(di);
                string lbtxt = (DialogueLines.Count - 1).ToString("0000") + ". ($" + (DialogueLines.Count).ToString("X4") + "). " + DialogueLines[DialogueLines.Count - 1].StrippedText;
                LB_Dialog.Items.Add(lbtxt);
                LB_Dialog.SelectedIndex = LB_Dialog.Items.Count - 1;
                remcaptions--;
                rembytes--;
                LBL_RemBytes.Text = rembytes.ToString();
                LBL_RemPointers.Text = remcaptions.ToString();
                NeedsSave(true);
            }
            else
            {
                MessageBox.Show("Maximum dialogue space reached.");
            }
        }

        private void BTN_Import_Click(object sender, EventArgs e)
        {
            OpenFileDialog ofd = new OpenFileDialog();
            ofd.ShowDialog();
            if (ofd.FileName != "")
            {
                DialogueLines.Clear();
                string filetext = File.ReadAllText(ofd.FileName);
                string[] captions = filetext.Replace("\r\n\r\n[caption #", "`").Split('`');
                for (int j = 1; j < captions.Length; j++)
                {
                    string cap = captions[j].Substring(captions[j].IndexOf("]\r\n") + "]\r\n".Length);
                    while (cap.EndsWith("\r\n"))
                    {
                        cap = cap.Substring(0, cap.LastIndexOf("\r\n"));
                    }
                    DialogueItem di = new DialogueItem(cap);
                    DialogueLines.Add(di);
                }

                LB_Dialog.Items.Clear();
                for (int i = 0; i < DialogueLines.Count; i++)
                {
                    string lbtxt = i.ToString("0000") + ". ($" + (i + 1).ToString("X4") + "). " + DialogueLines[i].StrippedText;
                    LB_Dialog.Items.Add(lbtxt);
                }
                NeedsSave(true);
            }
        }

        private void BTN_Export_Click(object sender, EventArgs e)
        {
            SaveFileDialog sfd = new SaveFileDialog();
            sfd.FileName = "dialog_script.txt";
            DialogResult dr = sfd.ShowDialog();
            if (dr == DialogResult.OK)
            {
                string exporttext = "[Script generated by: DialogueEditor, Coded by Fast Moon]\r\n" +
                    "[DialogueEditor version: 1.0.0]\r\n[File version: 1.00]\r\n[Section: Town Dialog]";
                for (int i = 0; i < DialogueLines.Count; i++)
                {
                    exporttext += "\r\n\r\n";
                    exporttext += "[caption #" + i.ToString() + "]\r\n";
                    exporttext += DialogueLines[i].Text;
                }
                exporttext += "\r\n\r\n";
                File.WriteAllText(sfd.FileName, exporttext);
            }
        }

        private void expandROMToolStripMenuItem_Click(object sender, EventArgs e)
        {
            DialogResult dr = MessageBox.Show("Expanding the ROM will add an additional megabyte of empty space to the end of the file that can hold expansions to the dialogue, events, maps, and other patches.\r\n\r\nWould you like to expand the ROM?", "Expand ROM", MessageBoxButtons.YesNo);
            if (dr == DialogResult.Yes)
            {
                byte[] expandeddata = new byte[0x400000];
                for(int i = 0; i < data.Length; i++)
                {
                    expandeddata[i] = data[i];
                }
                for(int j = 0x300000; j < 0x400000; j++)
                {
                    expandeddata[j] = 0xFF;
                }
                data = expandeddata;
                isROMExpanded = true;
                expandROMToolStripMenuItem.Enabled = false;
                expandROMToolStripMenuItem.Text = "ROM Already Expanded";
                expandDialogueSpaceToolStripMenuItem.Enabled = true;
                NeedsSave(true);
                MessageBox.Show("ROM data expanded.\r\n\r\nNo changes are saved to file until Save is selected from the menu.");
            }
        }

        private void expandDialogueSpaceToolStripMenuItem_Click(object sender, EventArgs e)
        {
            DialogResult dr = MessageBox.Show("Expanding the dialogue space will provide an additional bank (65KB) to store dialogue text.  By default the additional bank will be set to F0, however it and the original two banks can be moved to any other empty bank by selecting Manage Memory.\r\n\r\n" +
                "Note that the game's original text compression was not fully optimized and simply re-saving the original here will add about 7.8KB of text space automatically simply due to better compression.\r\n\r\nWould you like to expand the dialogue?", "Expand Dialogue", MessageBoxButtons.YesNo);
            if(dr == DialogResult.Yes)
            {
                byte[] dlgasmchange = { 0x7B, 0xAA, 0xC2, 0x20, 0xA5, 0xD0, 0xE8, 0xE8, 0xE8, 0xDF, 0x7D, 0xDF, 0xC0, 0xB0, 0xF7, 0x0A, 0x48, 0xBF, 0x7E, 0xDF, 0xC0, 0x85, 0xCA, 0xFA, 0xBF, 0x02, 0xE6, 0xCC, 0x85, 0xC9, 0xE2, 0x20, 0x7B, 0x1A, 0x8D, 0x68, 0x05, 0x60 };
                byte[] df80 = { data[0xCE600], data[0xCE601], 0xCD, 0xFF, 0xFF, 0xCE, 0xFF, 0xFF, 0xF0, 0xFE, 0x0C };
                for(int i = 0; i < dlgasmchange.Length; i++)
                {
                    data[0x7FBF + i] = dlgasmchange[i];
                }
                for(int j = 0; j < df80.Length; j++)
                {
                    data[0xDF80 + j] = df80[j];
                }
                isDialogExpanded = true;
                expandDialogueSpaceToolStripMenuItem.Enabled = false;
                expandDialogueSpaceToolStripMenuItem.Text = "Dialogue Already Expanded";
                manageMemoryToolStripMenuItem.Enabled = true;
                rembytes += 65536;
                LBL_RemBytes.Text = rembytes.ToString();
                NeedsSave(true);
                MessageBox.Show("Dialogue Expansion added.\r\n\r\nNo changes are saved to file until Save is selected from the menu.");
            }
        }
        
        public void NeedsSave(bool saveneeded)
        {
            if(saveneeded)
            {
                this.Text = ProgName + " -- Unsaved Changes --";
            }
            else
            {
                this.Text = ProgName + "FF3DE";
            }
        }

        private void BTN_DTERef_Click(object sender, EventArgs e)
        {
            DTEReference dtr = new DTEReference();
            dtr.Show();
        }
    }
}
