﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Globalization;
using System.Text.RegularExpressions;

namespace DialogueEditor
{
    class DialogueItem
    {
        string bytelist;
        string text;
        private string strippedtext;
        
        public DialogueItem(byte firstbyte)
        {
            bytelist = firstbyte.ToString("X2");
            text = "";
            strippedtext = "";
        }

        public DialogueItem(string txt)
        {
            text = FormatTextForMemory(txt);
            StripText(Bytes);
        }

        public string Text
        {
            get { return FormatTextForTextbox(text); }
            set
            {
                text = FormatTextForMemory(value);
                StripText(Bytes);
            }
        }

        public byte[] Bytes
        {
            get { return DTEConverter.BytesFromString(text); }
        }

        public string StrippedText
        {
            get { return strippedtext.Replace("\r\n", ""); }
        }

        private void StripText(byte[]bytearray)
        {
            int i = 0;
            text = "";
            strippedtext = "";
            while (i < bytearray.Length)
            {
                string nextchar = DTEConverter.StringFromByte(bytearray, i);
                text += nextchar;
                if (nextchar.Contains("<"))
                {
                    if (nextchar.StartsWith("<A"))
                    {
                        nextchar = nextchar.Replace("<A", "").Replace(">", "");
                        nextchar = DTEConverter.Names[int.Parse(nextchar)];
                        strippedtext += nextchar;
                    }
                }
                else
                {
                    strippedtext += nextchar;
                }
                i = i + DTEConverter.IndexAdvance;
            }
            strippedtext = strippedtext.Replace("_", "...").Replace("[", "\"").Replace("]", "\"");
        }

        public string FormatTextForMemory(string txt)
        {
            string newtxt = txt.Replace("\r", "").Replace("\n", "`");
            newtxt = newtxt.Replace("`", "\r\n");
            newtxt = newtxt.Replace("<EOP>\r\n", "<EOP>");
            int s = 0;
            string[] cmds = { "<E", "<OP", "<S", "<D" };
            foreach (string c in cmds)
            {
                while (newtxt.IndexOf(c, s) >= 0)
                {
                    s = newtxt.IndexOf(c, s);
                    int e = newtxt.IndexOf(">", s);
                    string cmd = newtxt.Substring(s, e - s + 1);
                    if(!(cmd.Contains("$") || cmd.Length == 3 || cmd.Contains("EO")))
                    {
                        string num = cmd.Replace(c, "").Replace(">", "");
                        string hexnum = "$" + int.Parse(num).ToString("X2");
                        string hexcmd = cmd.Replace(num, hexnum);
                        newtxt = newtxt.Replace(cmd, hexcmd);                        
                    }
                    s++;
                }
            }
            for(int i = 20; i > 2; i--)
            {
                newtxt = newtxt.Replace(new String(' ', i), "<S$" + i.ToString("X2") + ">");
            }
            if(!newtxt.EndsWith("<EOC>"))
            {
                newtxt += "<EOC>";
            }

            return newtxt;
        }

        public string FormatTextForTextbox(string txt)
        {
            string newtxt = Regex.Replace(txt, @"\<EOP\>(?!(\r\n|\r|\n))", "<EOP>\r\n");//txt.Replace(" < EOP>", "<EOP>\r\n");
            while(newtxt.IndexOf("<S$") >= 0)
            {
                string spacecount = newtxt.Substring(newtxt.IndexOf("<S$") + 3, 2);
                int spaces = int.Parse(spacecount, NumberStyles.AllowHexSpecifier);
                string space = new string(' ', spaces);
                newtxt = newtxt.Replace("<S$" + spacecount + ">", space);
            }
            newtxt = newtxt.Replace("<EOC>", "");
            return newtxt;
        }

        public void AddByte(byte b)
        {
            bytelist += "," + b.ToString("X2");
            if (b == 0)
            {
                string[] list = bytelist.Split(',');
                byte[] bytearray = Array.ConvertAll(list, item => byte.Parse(item, NumberStyles.AllowHexSpecifier));
                StripText(bytearray);
                /*int i = 0;
                while(i < bytearray.Length)
                {
                    string nextchar = DTEConverter.StringFromByte(bytearray, i);
                    text += nextchar;
                    if(nextchar.Contains("<"))
                    {
                        if(nextchar.StartsWith("<A"))
                        {
                            nextchar = nextchar.Replace("<A", "").Replace(">", "");
                            nextchar = DTEConverter.Names[int.Parse(nextchar)];
                            strippedtext += nextchar;
                        }
                    }
                    else
                    {
                        strippedtext += nextchar;
                    }
                    i = i + DTEConverter.IndexAdvance;
                }*/
            }
        }
    }

}
